@@ -1,12 +1,13 @@
 """ 
 Python SDK for LaunchKey API 
 For use in implementing LaunchKey
-Version 1.2.7
+Version 1.3.0
 @author LaunchKey
-@updated 2015-04-06
+@updated 2016-04-22
 """
 
-import requests
+import requests, six
+
 
 def generate_RSA(bits=2048):
     '''
@@ -19,7 +20,8 @@ def generate_RSA(bits=2048):
     new_key = RSA.generate(bits, e=65537)
     public_key = new_key.publickey().exportKey("PEM")
     private_key = new_key.exportKey("PEM")
-    return private_key, public_key
+    return private_key.decode('unicode_escape'), public_key.decode('unicode_escape')
+
 
 def decrypt_RSA(key, package):
     '''
@@ -34,7 +36,8 @@ def decrypt_RSA(key, package):
     rsakey = RSA.importKey(key)
     rsakey = PKCS1_OAEP.new(rsakey)
     decrypted = rsakey.decrypt(b64decode(package))
-    return decrypted
+    return decrypted.decode('unicode_escape')
+
 
 def encrypt_RSA(key, message):
     '''
@@ -45,11 +48,12 @@ def encrypt_RSA(key, message):
     '''
     from Crypto.PublicKey import RSA
     from Crypto.Cipher import PKCS1_OAEP
-    from base64 import b64encode
     rsakey = RSA.importKey(key)
     rsakey = PKCS1_OAEP.new(rsakey)
-    encrypted = rsakey.encrypt(message)
-    return encrypted.encode('base64')
+    encrypted = rsakey.encrypt(six.b(message))
+    from base64 import b64encode
+    return b64encode(encrypted).decode('unicode_escape')
+
 
 def sign_data(priv_key, data, encoded=True):
     '''
@@ -72,7 +76,8 @@ def sign_data(priv_key, data, encoded=True):
     else:
         digest.update(data)
     sign = signer.sign(digest)
-    return b64encode(sign)
+    return b64encode(sign).decode('unicode_escape')
+
 
 def verify_sign(pub_key, signature, data):
     '''
@@ -99,6 +104,7 @@ def verify_sign(pub_key, signature, data):
         pass
     return False
 
+
 def decrypt_AES(cipher, package, iv):
     '''
     Decrypts an AES package using a specific cipher and iv
@@ -143,7 +149,7 @@ class API(object):
         :param cipher: The AES cipher that was used to encrypt other parameters or body
         :return: Dict with RSA encrypted secret_key and signature of that value
         '''
-        #Ping to get key and time
+        # Ping to get key and time
         self.ping()
         to_encrypt = {"secret": self.app_secret, "stamped": str(self.ping_time)}
         if cipher is not None:
@@ -153,7 +159,7 @@ class API(object):
         if signature:
             signature = sign_data(self.private_key, encrypted_secret)
             to_return['signature'] = signature
-        
+
         to_return['app_key'] = self.app_key
         return to_return
 
@@ -181,22 +187,34 @@ class API(object):
             self.ping_time = datetime.datetime.now() - self.ping_difference + self.ping_time
         return {"launchkey_time": str(self.ping_time)[:-7], "key": self.api_pub_key}
 
-    def authorize(self, username, session=True, user_push_id=False):
+    def authorize(self, username, session=True, user_push_id=False, context=None, policy=None):
         '''
         Used to send an authorization request for a specific username
         :param username: String. The LaunchKey username of the one authorizing
         :param session: Boolean. If keeping a session mark True; transactional mark False
         :param user_push_id: Boolean. If your app would like to be returned an ID for the user
         that can be used to initiate notifications in the future without user input mark True.
+        :param context: String. Text the rocket would like to send along to the user about this
+        particular authentication request.
+        :param policy: Policy. Policy object to request a policy override for this request
         :return: String. The auth_request value for future reference.
         '''
         params = self._prepare_auth(signature=True)
         params['username'] = username
         params['session'] = session
         params['user_push_id'] = user_push_id
+        if context is not None:
+            params['context'] = context
+        if policy is not None:
+            import json
+            params['policy'] = json.dumps(policy)
         response = requests.post(self.API_HOST + "auths", params=params, verify=self.verify)
         try:
             if 'status_code' in response.json() and response.json()['status_code'] >= 300:
+                '''30421 - POST; Incorrect data for API call
+                30422 - POST; Credentials incorrect for app and app secret
+                30423 - POST; Error verifying app
+                30424 - POST; No paired devices'''
                 error = str(response.json()['message_code']) + " " + str(response.json()['message'])
                 return "Error: " + error
         except ValueError:
@@ -225,12 +243,11 @@ class API(object):
         import json
         auth_response = json.loads(decrypt_RSA(self.private_key, package))
         if "response" not in auth_response or "auth_request" not in auth_response or \
-                auth_request != auth_response['auth_request']:
+                        auth_request != auth_response['auth_request']:
             return self._notify("Authenticate", False, auth_request)
         response = str(auth_response['response']).lower() == 'true'
         return self._notify("Authenticate", response, auth_response['auth_request'])
 
-
     def _notify(self, action, status, auth_request):
         '''
         Notifies LaunchKey as to whether the user was logged in/out or not
@@ -252,7 +269,7 @@ class API(object):
         return False
 
     def deorbit(self, deorbit, signature):
-        ''' 
+        '''
         Verify the deorbit request by signature and timestamp 
         Return the user_hash needed to identify the user and log them out
         :param deorbit: JSON string from LaunchKey with the user_hash and
@@ -268,7 +285,7 @@ class API(object):
             decoded = json.loads(deorbit)
             date_request = datetime.datetime.strptime(decoded['launchkey_time'], "%Y-%m-%d %H:%M:%S")
             if self.ping_time - date_request < datetime.timedelta(minutes=5):
-                #Only want to honor a request that's been made recently
+                # Only want to honor a request that's been made recently
                 return decoded['user_hash']
         return None
 
@@ -286,7 +303,7 @@ class API(object):
         :return: user_hash
         '''
         raise NotImplementedError('Subclass must implement.')
-    
+
     def create_whitelabel_user(self, identifier):
         '''
         WhiteLabel Only
@@ -314,3 +331,53 @@ class API(object):
         data = decrypt_AES(cipher[:-16], response.json()['response']['data'], cipher[-16:])
         import json
         return json.loads(data)
+
+
+class Policy(dict):
+    """
+    Policy object for applying dynamic policies to requests
+    """
+    def __init__(self, all=0, knowledge=False, inherence=False, possession=False):
+        """
+        :param all: An integer number for the number of all tfactor types to require
+        :param knowledge: Flag to determine if a knowledge factor is required
+        :param inherence: Flag to determine if an inherence factor is required
+        :param possession: Flag to determine if a knowledge factor is required
+        """
+        if all and (knowledge or inherence or possession):
+            raise AttributeError(
+                "You must use either \"all\" or a factor (knowledge, inherence, or possession) not both"
+            )
+        self['minimum_requirements'] = [{
+            'requirement': 'authenticated',
+            'all': all,
+            'knowledge': 1 if knowledge else 0,
+            'inherence': 1 if inherence else 0,
+            'possession': 1 if possession else 0
+        }]
+        self['factors'] = []
+
+    def add_location(self, radius, latitude, longitude):
+        """
+        Add a location for geo-location restrictions on this policy
+        :param radius: Distance in meters
+        :param latitude: Degrees latitude
+        :param longitude: Degrees longitude
+        :return:
+        """
+        if not len(self['factors']):
+            self['factors'].append({
+                'factor': 'geofence',
+                'requirement': 'forced requirement',
+                'quickfail': True,
+                'priority': 1,
+                'attributes': {
+                    'locations': []
+                }
+
+            })
+        self['factors'][0]['attributes']['locations'].append({
+            "radius": radius,
+            "latitude": latitude,
+            "longitude": longitude
+        })
\ No newline at end of file
